/* gtkutil functions.
 */

/*

    Copyright (C) 1991-2003 The National Gallery

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

/*

    These files are distributed with VIPS - http://www.vips.ecs.soton.ac.uk

 */

#include "ip.h"

/*
#define DEBUG
 */

/* All our tooltips.
 */
static GtkTooltips *our_tooltips = NULL;

/* Set two adjustments together.
 */
void
adjustments_set_value( GtkAdjustment *hadj, GtkAdjustment *vadj,
        float hval, float vval )
{
        gboolean hchanged = FALSE;
        gboolean vchanged = FALSE;

        if( hval != hadj->value ) {
                hadj->value = hval;
                hchanged = TRUE;
        }
        if( vval != vadj->value ) {
                vadj->value = vval;
                vchanged = TRUE;
        }

#ifdef DEBUG
        if( hchanged )
                printf( "adjustments_set_value: hadj = %g\n", hval );
        if( vchanged )
                printf( "adjustments_set_value: vadj = %g\n", vval );
#endif /*DEBUG*/

        if( hchanged )
                gtk_adjustment_value_changed( hadj );
        if( vchanged )
                gtk_adjustment_value_changed( vadj );
}

void *
object_destroy( void *obj )
{
	gtk_object_destroy( GTK_OBJECT( obj ) );

	return( NULL );
}

/* Like g_free, but return NULL for list maps.
 */
void *
null_g_free( void *obj )
{
	g_free( obj );

	return( NULL );
}

/* Set visible/not.
 */
void
widget_visible( GtkWidget *widget, gboolean visible )
{
	if( widget && visible )
		gtk_widget_show( widget );
	else if( widget && !visible )
		gtk_widget_hide( widget );
}

/* Make a button widget.
 */
GtkWidget *
build_button( const char *stock_id, GtkSignalFunc cb, gpointer user )
{
	GtkWidget *but;

	but = gtk_button_new_from_stock( stock_id );
	GTK_WIDGET_SET_FLAGS( but, GTK_CAN_DEFAULT );
	gtk_signal_connect( GTK_OBJECT( but ), "clicked", cb, user );

	return( but );
}

/* Calculate the bounding box for a string rendered with a widget's default
 * font. Set geo to a rect with 0,0 positioned on the left-hand baseline.
 */
void
get_geo( GtkWidget *widget, const char *text, Rect *geo )
{
	PangoLayout *layout;
	int width, height;

	layout = gtk_widget_create_pango_layout( widget, text );
	pango_layout_get_pixel_size( layout, &width, &height );
	g_object_unref( layout );

	/* FIXME ... we left/top to 0 for now.
	 */
	geo->width = width;
	geo->height = height;
	geo->left = 0;
	geo->top = 0;
}

/* Set a widget to a fixed size ... width in characters.
 */
void
set_fixed( GtkWidget *widget, int nchars )
{
	Rect geo;

	get_geo( widget, "8", &geo );
	gtk_widget_set_size_request( widget, geo.width * nchars, geo.height );
}

/* Build a GtkEntry, with a widget width specified in characters.
 */
GtkWidget *
build_entry( int nchars )
{
	GtkWidget *entry;

	entry = gtk_entry_new();
	gtk_entry_set_width_chars( GTK_ENTRY( entry ), nchars );

	return( entry );
}

/* Build a new menu.
 */
GtkWidget *
menu_build( const char *name )
{
	GtkWidget *menu;

	menu = gtk_menu_new();
	gtk_menu_set_title( GTK_MENU( menu ), name );

	return( menu );
}

/* Add a menu item.
 */
GtkWidget *
menu_add_but( GtkWidget *menu, 
	const char *stock_id, GtkSignalFunc cb, void *user )
{
	GtkWidget *but;

	/* We don't provide an accel group for popup menus.
	 */
	but = gtk_image_menu_item_new_from_stock( stock_id, NULL );
	gtk_menu_shell_append( GTK_MENU_SHELL( menu ), but );
	gtk_widget_show( but );
	gtk_signal_connect( GTK_OBJECT( but ), "activate", cb, user );

	return( but );
}

/* Add a toggle item.
 */
GtkWidget *
menu_add_tog( GtkWidget *menu, const char *name, GtkSignalFunc cb, void *user )
{
	GtkWidget *tog;

	tog = gtk_check_menu_item_new_with_mnemonic( name );
	gtk_menu_shell_append( GTK_MENU_SHELL( menu ), tog );
	gtk_widget_show( tog );
	gtk_signal_connect( GTK_OBJECT( tog ), "toggled", cb, user );

	return( tog );
}

/* Add a separator.
 */
GtkWidget *
menu_add_sep( GtkWidget *menu )
{
	GtkWidget *sep;

	sep = gtk_menu_item_new();
	gtk_widget_set_sensitive( GTK_WIDGET( sep ), FALSE );
	gtk_menu_shell_append( GTK_MENU_SHELL( menu ), sep );
	gtk_widget_show( sep );

	return( sep );
}

/* Four quarks: each menu item has a quark linking back to the main pane, 
 * plus a quark for the user signal. The main pane has a quark linking to the
 * widget the menu was popped from, and that has the userdata for this context.
 * One more quark holds the popup in a host.
 */
static GQuark quark_main = 0;
static GQuark quark_host = 0;
static GQuark quark_data = 0;
static GQuark quark_popup = 0;

/* Build a new popup menu.
 */
GtkWidget *
popup_build( const char *name )
{
	/* Build our quarks.
	 */
	if( !quark_main ) {
		quark_main = g_quark_from_static_string( "quark_main" );
		quark_host = g_quark_from_static_string( "quark_host" );
		quark_data = g_quark_from_static_string( "quark_data" );
		quark_popup = g_quark_from_static_string( "quark_popup" );
	}

	return( menu_build( name ) );
}

/* Activate function for a popup menu item.
 */
static void
popup_activate_cb( GtkWidget *item, PopupFunc cb )
{
	GtkWidget *qmain = gtk_object_get_data_by_id( 
		GTK_OBJECT( item ), quark_main );
	GtkWidget *qhost = gtk_object_get_data_by_id( 
		GTK_OBJECT( qmain ), quark_host );
	void *qdata = gtk_object_get_data_by_id( 
		GTK_OBJECT( qhost ), quark_data );

	(*cb)( item, qhost, qdata );
}

/* Add a menu item to a popup.
 */
GtkWidget *
popup_add_but( GtkWidget *popup, const char *name, PopupFunc cb )
{
	GtkWidget *but = menu_add_but( popup, name, 
		GTK_SIGNAL_FUNC( popup_activate_cb ), (void *) cb );

	gtk_object_set_data_by_id( GTK_OBJECT( but ), quark_main, popup );

	return( but );
}

/* Add a toggle item to a popup.
 */
GtkWidget *
popup_add_tog( GtkWidget *popup, const char *name, PopupFunc cb )
{
	GtkWidget *tog = menu_add_tog( popup, name, 
		GTK_SIGNAL_FUNC( popup_activate_cb ), (void *) cb );

	gtk_object_set_data_by_id( GTK_OBJECT( tog ), quark_main, popup );

	return( tog );
}

/* Show the popup.
 */
void
popup_show( GtkWidget *host, GdkEvent *ev )
{
	GtkWidget *popup = gtk_object_get_data_by_id( 
		GTK_OBJECT( host ), quark_popup );

	gtk_object_set_data_by_id( GTK_OBJECT( popup ), quark_host, host );
	gtk_menu_popup( GTK_MENU( popup ), NULL, NULL,
		(GtkMenuPositionFunc) NULL, NULL, 3, ev->button.time );
}

/* Event handler for popupshow.
 */
static gboolean
popup_handle_event( GtkWidget *host, GdkEvent *ev, gpointer dummy )
{
	gboolean handled = FALSE;

        if( ev->type == GDK_BUTTON_PRESS && ev->button.button == 3 ) {
		popup_show( host, ev );
		handled = TRUE;
	}
	else if( ev->type == GDK_KEY_PRESS && ev->key.keyval == GDK_F10 && 
		ev->key.state & GDK_SHIFT_MASK ) {
		popup_show( host, ev );
		handled = TRUE;
	}

	return( handled );
}

/* Link a host to a popup.
 */
void
popup_link( GtkWidget *host, GtkWidget *popup, void *data )
{
	gtk_object_set_data_by_id( GTK_OBJECT( host ), quark_popup, popup );
	gtk_object_set_data_by_id( GTK_OBJECT( host ), quark_data, data );
}

/* Add a callback to show a popup.
 */
guint
popup_attach( GtkWidget *host, GtkWidget *popup, void *data )
{
	guint sid;

	popup_link( host, popup, data );
        sid = gtk_signal_connect( GTK_OBJECT( host ), "event",
                GTK_SIGNAL_FUNC( popup_handle_event ), NULL );

	return( sid );
}

void
popup_detach( GtkWidget *host, guint sid )
{
	gtk_signal_disconnect( GTK_OBJECT( host ), sid );
}

static void
set_tooltip_events( GtkWidget *wid )
{
	gtk_widget_add_events( wid, 
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK );
}

/* Set the tooltip on a widget.
 */
void
set_tooltip( GtkWidget *wid, const char *fmt, ... )
{
	va_list ap;
	char *txt;

	if( !wid )
		return;

	if( !fmt )
		fmt = "";

	va_start( ap, fmt );
	txt = g_strdup_vprintf( fmt, ap );
	va_end( ap );

	if( !our_tooltips ) 
		our_tooltips = gtk_tooltips_new();

	gtk_tooltips_set_tip( our_tooltips, wid, txt, NULL );

	if( !GTK_WIDGET_REALIZED( wid ) )
		gtk_signal_connect( GTK_OBJECT( wid ), "realize",
			GTK_SIGNAL_FUNC( set_tooltip_events ), NULL );
	else
		set_tooltip_events( wid );

	g_free( txt );
}

/* Set a GtkEditable.
 */
void
set_gentryv( GtkWidget *edit, const char *fmt, va_list ap )
{
	char buf[1000];
	gint position;
	int i;
	int len;

	if( !edit )
		return;

	if( !fmt )
		fmt = "";

	(void) im_vsnprintf( buf, 1000, fmt, ap );

	/* Filter out /n and /t ... they confuse gtkentry terribly
	 */
	len = strlen( buf );
	for( i = 0; i < len; i++ )
		if( buf[i] == '\n' || buf[i] == '\t' )
			buf[i] = ' ';

	gtk_editable_delete_text( GTK_EDITABLE( edit ), 0, -1 );
	position = 0;
	gtk_editable_insert_text( GTK_EDITABLE( edit ), 
		buf, strlen( buf ), &position );
}

/* Set a GtkEditable.
 */
void
set_gentry( GtkWidget *edit, const char *fmt, ... )
{
	va_list ap;

	va_start( ap, fmt );
	set_gentryv( edit, fmt, ap );
	va_end( ap );
}

void
set_glabel( GtkWidget *label, const char *fmt, ... )
{
	va_list ap;
	char buf[1000];

	va_start( ap, fmt );
	(void) im_vsnprintf( buf, 1000, fmt, ap );
	va_end( ap );

	gtk_label_set_text( GTK_LABEL( label ), buf );
}

/* Like set_glabel, but do it caption-style.
 */
void 
set_gcaption( GtkWidget *label, const char *fmt, ... )
{
	va_list ap;
	char buf1[1000];
	char buf2[1000];

	va_start( ap, fmt );
	(void) im_vsnprintf( buf1, 1000, fmt, ap );
	va_end( ap );

	escape_markup( buf1, buf2, 1000 );
	(void) im_snprintf( buf1, 1000, 
		"<span size=\"smaller\">%s</span>", buf2 );
	gtk_label_set_markup( GTK_LABEL( label ), buf1 );
}

gboolean
get_geditable_name( GtkWidget *text, char *out, int sz )
{
	char *name;
	char *tname;

	name = gtk_editable_get_chars( GTK_EDITABLE( text ), 0, -1 );
	tname = trim_nonalpha( name );
	if( !tname ) {
		IM_FREEF( g_free, name );
		error_top( _( "Bad identifier." ) );
		error_top( 
			_( "Enter an identifier. Identifiers start with "
			"a letter, and then contain only letters, numbers, "
			"apostrophy and underscore." ) );
		return( FALSE );
	}
	im_strncpy( out, tname, sz );
	IM_FREEF( g_free, name );

	return( TRUE );
}

gboolean
get_geditable_string( GtkWidget *text, char *out, int sz )
{
	char *str;

	str = gtk_editable_get_chars( GTK_EDITABLE( text ), 0, -1 );
	im_strncpy( out, str, sz );
	IM_FREEF( g_free, str );

	return( TRUE );
}

gboolean
get_geditable_filename( GtkWidget *text, char *out, int sz )
{
	char *filename;
	char *tfilename;

	filename = gtk_editable_get_chars( GTK_EDITABLE( text ), 0, -1 );
	tfilename = trim_white( filename );
	if( !is_valid_filename( tfilename ) ) {
		IM_FREEF( g_free, filename );
		return( FALSE );
	}
	im_strncpy( out, tfilename, sz );
	IM_FREEF( g_free, filename );

	return( TRUE );
}

/* Get a geditable as a double.
 */
gboolean
get_geditable_double( GtkWidget *text, double *out )
{
	char *txt = gtk_editable_get_chars( GTK_EDITABLE( text ), 0, -1 );
	char *end;
	double t;

	t = strtod( txt, &end );
	if( end == txt ) {
		error_top( _( "Bad floating point number." ) );
		error_sub( _( "\"%s\" is not a floating point number." ), txt );
		IM_FREEF( g_free, txt );
		return( FALSE );
	}

	if( strspn( end, WHITESPACE ) != strlen( end ) ) {
		error_top( _( "Bad floating point number." ) );
		error_sub( _( "Extra characters \"%s\" after number." ), end );
		IM_FREEF( g_free, txt );
		return( FALSE );
	}

	IM_FREEF( g_free, txt );
	*out = t;

	return( TRUE );
}

/* Get as int.
 */
gboolean
get_geditable_int( GtkWidget *text, int *n )
{
	int i;
	char *txt = gtk_editable_get_chars( GTK_EDITABLE( text ), 0, -1 );

	/* Parse values.
	 */
	if( sscanf( txt, "%i", &i ) != 1 ) {
		error_top( _( "Bad integer." ) );
		error_sub( _( "\"%s\" is not an integer." ), txt );
		IM_FREEF( g_free, txt );
		return( FALSE );
	}
	IM_FREEF( g_free, txt );
	*n = i;

	return( TRUE );
}

/* Get as unsigned int.
 */
gboolean
get_geditable_uint( GtkWidget *text, int *n )
{
	int i;

	if( !get_geditable_int( text, &i ) || i < 0 ) {
		error_top( _( "Bad unsigned integer." ) );
		return( FALSE );
	}
	*n = i;

	return( TRUE );
}

/* Get as positive int.
 */
gboolean
get_geditable_pint( GtkWidget *text, int *n )
{
	int i;

	if( !get_geditable_int( text, &i ) || i <= 0 ) {
		error_top( _( "Bad positive integer." ) );
		return( FALSE );
	}
	*n = i;

	return( TRUE );
}

/* Set a name for a widget and all it's children ... so they pick up a style.
 */
void 
set_name( GtkWidget *widget, const char *name )
{
	gtk_widget_set_name( widget, name );
	if( GTK_IS_CONTAINER( widget ) )
		gtk_container_foreach( GTK_CONTAINER( widget ), 
                           (GtkCallback) set_name, (char *) name );
}

/* Indent widget, label above.
 */
GtkWidget *
build_glabelframe2( GtkWidget *widget, const char *name )
{
	GtkWidget *lab;
	GtkWidget *vb;
	GtkWidget *hb;
	GtkWidget *inv;
	char buf[1000];

        hb = gtk_hbox_new( FALSE, 2 );
        inv = gtk_label_new( "" );
        gtk_box_pack_start( GTK_BOX( hb ), inv, FALSE, FALSE, 15 );
        gtk_box_pack_start( GTK_BOX( hb ), widget, TRUE, TRUE, 0 );

        vb = gtk_vbox_new( FALSE, 2 );
	im_snprintf( buf, 1000, _( "%s:" ), name );
	lab = gtk_label_new( buf );
	gtk_misc_set_alignment( GTK_MISC( lab ), 0.0, 0.5 );
        gtk_box_pack_start( GTK_BOX( vb ), lab, FALSE, FALSE, 0 );
        gtk_box_pack_start( GTK_BOX( vb ), hb, TRUE, TRUE, 0 );

	return( vb );
}

/* Make a text field + label. Indent the text on a new line.
 */
GtkWidget *
build_glabeltext3( GtkWidget *box, const char *label )
{	
	GtkWidget *txt;
	GtkWidget *vb;

        txt = gtk_entry_new();
        vb = build_glabelframe2( txt, label );
        gtk_box_pack_start( GTK_BOX( box ), vb, FALSE, FALSE, 0 );

	return( txt );
}

/* Make text field plus label .. use a sizegroup for alignment.
 */
GtkWidget *
build_glabeltext4( GtkWidget *box, GtkSizeGroup *group, const char *text )
{
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *entry;
	char buf[256];

	hbox = gtk_hbox_new( FALSE, 12 );
	im_snprintf( buf, 256, _( "%s:" ), text );
	label = gtk_label_new( buf );
	gtk_misc_set_alignment( GTK_MISC( label ), 0.0, 0.5 );
	if( group )
		gtk_size_group_add_widget( group, label );  
	gtk_box_pack_start( GTK_BOX( hbox ), label, FALSE, FALSE, 0 );

	entry = gtk_entry_new();
	gtk_box_pack_start( GTK_BOX( hbox ), entry, TRUE, TRUE, 0 );

	gtk_box_pack_start( GTK_BOX( box ), hbox, FALSE, FALSE, 0 );

	gtk_widget_show_all( hbox );

	return( entry );
}

/* Make a labeled toggle.
 */
GtkWidget *
build_gtoggle( GtkWidget *box, const char *caption )
{
        GtkWidget *hb;
        GtkWidget *inv;
	GtkWidget *toggle;

	/* Indent left a bit.
	 */
        inv = gtk_label_new( "" );
        hb = gtk_hbox_new( FALSE, 0 );
        gtk_box_pack_start( GTK_BOX( hb ), inv, FALSE, FALSE, 2 );
        toggle = gtk_check_button_new_with_label( caption );
	gtk_container_set_border_width( GTK_CONTAINER( toggle ), 4 );
        gtk_box_pack_start( GTK_BOX( hb ), toggle, TRUE, TRUE, 0 );

        gtk_box_pack_start( GTK_BOX( box ), hb, FALSE, FALSE, 0 );

	return( toggle );
}

/* Make a label plus option menu.
 */
GtkWidget *
build_goption( GtkWidget *box, GtkSizeGroup *group, 
	const char *name, const char *item_names[], int nitem,
	GtkSignalFunc fn, void *value )
{
	GtkWidget *hb;
	GtkWidget *label;
	GtkWidget *om;
	int i;
	char buf[1000];

        hb = gtk_hbox_new( FALSE, 12 );
	im_snprintf( buf, 1000, _( "%s:" ), name );
        label = gtk_label_new( buf );
	if( group )
		gtk_size_group_add_widget( group, label );  
        gtk_box_pack_start( GTK_BOX( hb ), label, FALSE, TRUE, 0 );

	om = gtk_combo_box_new_text();
        gtk_box_pack_start( GTK_BOX( hb ), om, FALSE, TRUE, 0 );
        set_tooltip( om, _( "Left-click to change value" ) );

	for( i = 0; i < nitem; i++ ) 
		gtk_combo_box_append_text( GTK_COMBO_BOX( om ),
			 _( item_names[i] ) );
	if( fn ) 
		gtk_signal_connect( GTK_OBJECT( om ), "changed", fn, value );
        gtk_box_pack_start( GTK_BOX( box ), hb, FALSE, TRUE, 0 );
        gtk_widget_show_all( hb );

	return( om );
}

/* Register a widget as a filename drag receiver.
 */
typedef struct {
	GtkWidget *widget;
	FiledropFunc fn;
	void *client;
} FiledropInfo;

static void
filedrop_drag_data_received( GtkWidget *widget, 
	GdkDragContext *context,
	gint x, gint y, GtkSelectionData *data, guint info, guint time,
	FiledropInfo *fdi )
{
	gchar *sPath = NULL;
	gchar *pFrom, *pTo; 
	gboolean result;

        pFrom = strstr( (char *) data->data, "file:" );

        while( pFrom ) {
#if !GLIB_CHECK_VERSION (2,0,0)
		pFrom += 5; /* remove 'file:' */
#else
		GError *error = NULL;
#endif

		pTo = pFrom;
		while( *pTo != 0 && *pTo != 0xd && *pTo != 0xa ) 
			pTo += 1;

		sPath = g_strndup( pFrom, pTo - pFrom );

#if !GLIB_CHECK_VERSION (2,0,0)
		result = fdi->fn( fdi->client, sPath );
#else
		/* format changed with Gtk+1.3, use conversion */
		pFrom = g_filename_from_uri( sPath, NULL, &error );
		result = fdi->fn( fdi->client, pFrom );
		g_free( pFrom );
#endif

		g_free( sPath );

		if( !result )
			box_alert( fdi->widget );

		pFrom = strstr( pTo, "file:" );
        } 

        gtk_drag_finish( context, TRUE, FALSE, time );
}

/* HB: file dnd stuff lent by The Gimp via Dia, not fully understood 
 * but working ...
 */
enum {
	TARGET_URI_LIST,
	TARGET_TEXT_PLAIN
};

static void
filedrop_destroy( GtkWidget *widget, FiledropInfo *fdi )
{
	im_free( fdi );
}

void
filedrop_register( GtkWidget *widget, FiledropFunc fn, void *client )
{
	static GtkTargetEntry target_table[] = {
		{ "text/uri-list", 0, TARGET_URI_LIST },
		{ "text/plain", 0, TARGET_TEXT_PLAIN }
	};

	FiledropInfo *fdi = INEW( NULL, FiledropInfo );

	fdi->widget = widget;
	fdi->fn = fn;
	fdi->client = client;
	gtk_signal_connect( GTK_OBJECT( widget ), "destroy",
                GTK_SIGNAL_FUNC( filedrop_destroy ), fdi );

	gtk_drag_dest_set( GTK_WIDGET( widget ),
		GTK_DEST_DEFAULT_ALL,
		target_table, IM_NUMBER( target_table ),
		GDK_ACTION_COPY );
	gtk_signal_connect( GTK_OBJECT( widget ), "drag_data_received",
		GTK_SIGNAL_FUNC( filedrop_drag_data_received ), fdi );
}

typedef struct _Listen {
	GObject *gobject;		/* This object */
	GObject *source;		/* Listens for signals from this */
	GObject **zap;			/* NULL this on destroy */
	const char *name;		/* Signal name */
	GCallback gcallback;		/* Call this handler */

	guint name_sid;
	guint gobject_destroy_sid;
	guint source_destroy_sid;
} Listen;

static void
listen_gobject_destroy_cb( GObject *gobject, Listen *listen )
{
	/* gobject has gone ... source should no longer send us signals.
	 */
	FREESID( listen->name_sid, listen->source );
	FREESID( listen->source_destroy_sid, listen->source );

	g_free( listen );
}

static void
listen_source_destroy_cb( GObject *gobject, Listen *listen )
{
	/* Source has gone, these signals have been destroyed.
	 */
	listen->name_sid = 0;
	listen->source_destroy_sid = 0;

	/* Link broken, no need to auto-free us on gobject destroy.
	 */
	FREESID( listen->gobject_destroy_sid, listen->gobject );

	/* Zap gobject member pointer to source.
	 */
	if( listen->zap ) {
		assert( !*(listen->zap) || *(listen->zap) == listen->source );

		*(listen->zap) = NULL;
	}

	g_free( listen );
}

void
listen_add( GObject *gobject, GObject **zap, 
	const char *name, GCallback gcallback )
{
	Listen *listen = g_new( Listen, 1 );

	listen->gobject = gobject;
	listen->source = *zap;
	listen->zap = zap;
	listen->name = name;
	listen->gcallback = gcallback;

	listen->name_sid = g_signal_connect( listen->source, 
		listen->name, listen->gcallback, listen->gobject );
	listen->source_destroy_sid = g_signal_connect( listen->source, 
		"destroy",
		G_CALLBACK( listen_source_destroy_cb ), listen );
	listen->gobject_destroy_sid = g_signal_connect( gobject, "destroy",
		G_CALLBACK( listen_gobject_destroy_cb ), listen );
}

void
widget_update_pointer( GtkWidget *widget, GdkEvent *ev )
{
	if( ev->type == GDK_MOTION_NOTIFY && ev->motion.is_hint ) {
		GdkDisplay *display = gtk_widget_get_display( widget );
		GdkScreen *screen;
		int x_root, y_root;

		gdk_display_get_pointer( display, 
			&screen, &x_root, &y_root, NULL );
		ev->motion.x_root = x_root;
		ev->motion.y_root = y_root; 
	}
}

void *
gobject_print( GObject *gobject )
{
	printf( "%s (0x%x)\n", 
		G_OBJECT_TYPE_NAME( gobject ), GPOINTER_TO_UINT( gobject ) );

	return( NULL );
}

/* Get the default DPI.
 */
int
get_dpi( void )
{
	GdkScreen *screen = gdk_screen_get_default();

	if( screen ) {
		int width_pixels = gdk_screen_get_width( screen );
		int width_mm = gdk_screen_get_width_mm( screen );

		return( width_pixels / (width_mm / 25.4) );
	}
	else
		return( 72 );
}

GtkWidget *
image_new_from_file( const char *name )
{
	GtkWidget *image;
	char *file;

	if( (file = path_find_file( PATH_SEARCH, name )) ) {
		image = (GtkWidget *) call_string_filename( 
			(call_string_fn) gtk_image_new_from_file, 
			file, NULL, NULL, NULL );
		im_free( file );
	}
	else
		/* We get a broken image icon if this fails.
		 */
		image = gtk_image_new_from_file( name );

	return( image );
}
